try:
    import typing 
except ImportError:
    from typing_extensions import * 
import cxxjij
import openjij.model
import openjij.sampler 
import openjij.sampler.chimera_gpu
import openjij.utils

from openjij.variable_type import SPIN, BINARY, Vartype, cast_vartype
from openjij.sampler import Response
from openjij.sampler import SASampler, SQASampler, CSQASampler
from openjij.model import BinaryQuadraticModel, BinaryPolynomialModel
from openjij.utils import solver_benchmark, convert_response



# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cxxjij
import openjij
from openjij.sampler import SASampler
from openjij.model import BinaryQuadraticModel, ChimeraModel
from .base_gpu_chimera import BaseGPUChimeraSampler
from openjij.utils.graph_utils import chimera_to_ind
import numpy as np
import dimod


class GPUChimeraSASampler(SASampler, BaseGPUChimeraSampler):
    """Sampler with Simulated Annealing (SA) on GPU.

    Inherits from :class:`openjij.sampler.sampler.BaseSampler`.

    Args:
        beta_min (float): Minimum inverse temperature.
        beta_max (float): Maximum inverse temperature.
        num_sweeps (int): Length of Monte Carlo step.
        schedule_info (dict): Information about a annealing schedule.
        num_reads (int): Number of iterations.
        unit_num_L (int): Length of one side of two-dimensional lattice in which chimera unit cells are arranged.

    Raises:
        ValueError: If variables violate as below.
        - trotter number is odd.
        - no input "unit_num_L" to an argument or this constructor.
        - given problem graph is incompatible with chimera graph.

        AttributeError: If GPU doesn't work.

    """

    def __init__(self,
                 beta_min=None, beta_max=None,
                 num_sweeps=1000, schedule=None,
                 num_reads=1, unit_num_L=None,
                 ):

        super().__init__(beta_min=beta_min, beta_max=beta_max, 
                         num_reads=num_reads, num_sweeps=num_sweeps, 
                         schedule=schedule)

        self.unit_num_L = unit_num_L

        self._make_system = {
            'singlespinflip': cxxjij.system.make_chimera_classical_gpu
        }
        self._algorithm = {
            'singlespinflip': cxxjij.algorithm.Algorithm_GPU_run
        }

    def sample_ising(self, h, J, beta_min=None, beta_max=None,
                     num_sweeps=None, num_reads=1, schedule=None,
                     initial_state=None, updater=None,
                     reinitialize_state=True, seed=None, unit_num_L=None,
                     ):
        """sample with Ising model.

        Args:
            h (dict): linear biases
            J (dict): quadratic biases
            beta_min (float): minimal value of inverse temperature
            beta_max (float): maximum value of inverse temperature
            num_sweeps (int): number of sweeps
            num_reads (int): number of reads
            schedule (list): list of inverse temperature
            initial_state (dict): initial state
            updater(str): updater algorithm
            reinitialize_state (bool): if true reinitialize state for each run
            seed (int): seed for Monte Carlo algorithm
            unit_num_L (int): number of chimera units

        Returns:
            :class:`openjij.sampler.response.Response`: results

        Examples::
            
            >>> sampler = oj.GPUChimeraSASampler(unit_num_L=2)
            >>> h = {0: -1, 1: -1, 2: 1, 3: 1},
            >>> J = {(0, 4): -1, (2, 5): -1}
            >>> res = sampler.sample_ising(h, J)

        """

        if updater is None:
            updater='single spin flip'
        

        self.unit_num_L = unit_num_L if unit_num_L else self.unit_num_L

        model = openjij.ChimeraModel(linear=h, quadratic=J, vartype='SPIN', 
                                   unit_num_L=self.unit_num_L, gpu=True)

        # define Chimera structure
        structure = {}
        structure['size'] = 8 * self.unit_num_L * self.unit_num_L
        structure['dict'] = {}
        if isinstance(model.indices[0], int):
            # identity dict
            for ind in model.indices:
                structure['dict'][ind] = ind
        elif isinstance(model.indices[0], tuple):
            # map chimera coordinate to index
            for ind in model.indices:
                structure['dict'][ind] = model.to_index(*ind, self.unit_num_L)
            

        return self._sampling(model, beta_min, beta_max,
                              num_sweeps, num_reads, schedule,
                              initial_state, updater,
                              reinitialize_state, seed, structure)

        


    




import functools
import warnings


def deprecated_alias(**aliases):
    def deco(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)
        return wrapper
    return deco


def rename_kwargs(func_name, kwargs, aliases):
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise TypeError('{} received both {} and {}'.format(
                    func_name, alias, new))
            warnings.warn('{} is deprecated; use {}'.format(alias, new),
                          DeprecationWarning)
            kwargs[new] = kwargs.pop(alias)


def disabled(func):
    def wrapper(*args, **kwargs):
        raise NotImplementedError("The function {} is disabled.".format(func.__name__))

    return wrapper



# import openjij as oj
# import cxxjij as cj
# import numpy as np
# 
# import unittest
# 
# class TestGPUModel(unittest.TestCase):
#     def setUp(self):
#         self.h = {0: 1, 1: -2}
#         self.J = {(0, 1): -1, (1, 2): -3, (2, 3): 0.5}
#         self.spins = {0: 1, 1: -1, 2: 1, 3: 1}
# 
#         self.Q = {(0, 0): 1, (1, 2): -1, (2, 0): -0.2, (1, 3): 3}
#         self.binaries = {0: 0, 1: 1, 2: 1, 3: 0}
# 
#     def test_transfer_to_cxxjij(self):
#         bqm = oj.BinaryQuadraticModel(self.h, self.J, gpu=True)
#         # to Dense
#         ising_graph = bqm.get_cxxjij_ising_graph(sparse=False)
#         self.assertTrue(isinstance(ising_graph, cj.graph.DenseGPU))
#         self.assertEqual(ising_graph.size(), len(bqm.indices))
#         for i in range(len(bqm.indices)):
#             for j in range(len(bqm.indices)):
#                 if i != j:
#                     self.assertAlmostEqual(bqm.interaction_matrix()[i,j], ising_graph.get_interactions()[i, j])
#                 else:
#                     # i == j
#                     self.assertAlmostEqual(bqm.interaction_matrix()[i,j], ising_graph.get_interactions()[i, len(bqm.indices)])
#                     self.assertAlmostEqual(bqm.interaction_matrix()[i,j], ising_graph.get_interactions()[len(bqm.indices), i])
#                     self.assertEqual(ising_graph.get_interactions()[i,i], 0)
# 
#         self.assertEqual(ising_graph.get_interactions()[len(bqm.indices),len(bqm.indices)], 1)
# 
# 
#         # to Sparse
#         ising_graph = bqm.get_cxxjij_ising_graph(sparse=True)
#         self.assertTrue(isinstance(ising_graph, cj.graph.SparseGPU))
#         self.assertEqual(ising_graph.size(), len(bqm.indices))
#         for i in range(ising_graph.size()):
#             for j in ising_graph.adj_nodes(i):
#                 self.assertEqual(bqm.interaction_matrix()[i,j], ising_graph[i,j])
# 
# 
# 
# class TestGPUSampler(unittest.TestCase):
#     def setUp(self):
#         self.num_ind = {
#             'h': {0: -1, 1: -1, 2: 1, 3: 1},
#             'J': {(0, 4): -1, (2, 5): -1}
#         }
#         str_ind = ['a', 'b', 'c', 'd', 'e', 'd']
#         self.str_ising = {
#             'h': {str_ind[i]: v for i, v in self.num_ind['h'].items()},
#             'J': {(str_ind[i], str_ind[j]): v for (i, j), v in self.num_ind['J'].items()}
#         }
#         self.ground_state = [1, 1, -1, -1, 1, -1]
#         self.e_g = -1-1-1-1 + (-1-1)
#         # self.g_sample = {i: self.ground_state[i]
#         #                  for i in range(len(self.ground_state))}
#         # self.g_samp_str = {k: self.ground_state[i]
#         #                    for i, k in enumerate(str_ind)}
# 
#         self.qubo = {
#             (0, 0): -1, (1, 1): -1, (2, 2): -1, (3, 3): 1,
#             (0, 4): -1, (2, 5): -1
#         }
#         # self.str_qubo = {(str_ind[i], str_ind[j]): qij
#         #                  for (i, j), qij in self.qubo.items()}
#         self.ground_q = [1, 1, 1, 0, 1, 1]
#         self.e_q = -1-1-1 + (-1-1)
# 
#     def samplers(self, sampler, init_state=None):
#         res = sampler.sample_ising(
#             self.num_ind['h'], self.num_ind['J'],
#             initial_state=init_state, seed=1)
#         self._test_response(res, self.e_g, self.ground_state)
#         res = sampler.sample_qubo(self.qubo,
#                                   initial_state=init_state, seed=1)
#         
#         self._test_response(res, self.e_q, self.ground_q)
# 
#         # embedding composit
#         # from dwave.system.composites import EmbeddingComposite
#         # samp = EmbeddingComposite(sampler)
#         # res = samp.sample_ising(
#         #     self.num_ind['h'], self.num_ind['J']
#         # )
#         # 
#         # res = samp.sample_ising(
#         #     self.str_ising['h'], self.str_ising['J']
#         # )
# 
#     def _test_response(self, res, e_g, s_g):
#         # test openjij response interface
#         self.assertEqual(len(res.states), 1)
#         self.assertListEqual(s_g, list(res.states[0]))
#         self.assertEqual(res.energies[0], e_g)
#         # test dimod interface
#         self.assertEqual(len(res.record.sample), 1)
#         self.assertListEqual(s_g, list(res.record.sample[0]))
#         self.assertEqual(res.record.energy[0], e_g)
# 
#     def _test_response_num(self, res, num_reads):
#         # test openjij response interface
#         self.assertEqual(len(res.states), num_reads)
#         self.assertEqual(len(res.energies), num_reads)
#         # test dimod interface
#         self.assertEqual(len(res.record.sample), num_reads)
#         self.assertEqual(len(res.record.energy), num_reads)
# 
#     def _test_num_reads(self, sampler_cls):
#         num_reads = 10
#         sampler = sampler_cls()
#         res = sampler.sample_ising(
#             self.num_ind['h'], self.num_ind['J'],
#             num_reads=num_reads,
#             seed=1
#         )
#         self._test_response_num(res, num_reads)
# 
#         sampler = sampler_cls(num_reads=num_reads)
#         res = sampler.sample_ising(
#             self.num_ind['h'], self.num_ind['J'],
#         )
#         self._test_response_num(res, num_reads)
# 
#     def test_sa(self):
#         sampler = oj.GPUChimeraSASampler(unit_num_L=2)
#         self.samplers(sampler)
#         self._test_num_reads(oj.SASampler)
# 
#     def test_sqa(self):
#         sampler = oj.GPUChimeraSQASampler(unit_num_L=2)
#         self.samplers(sampler)
#         self._test_num_reads(oj.SQASampler)
# 
# 
# 
# if __name__ == '__main__':
#     unittest.main()



import numpy as np
import openjij
from openjij.sampler import measure_time
from openjij.sampler import SQASampler
from openjij.utils.decorator import deprecated_alias
import cxxjij


class CSQASampler(SQASampler):
    """Sampler with continuous-time simulated quantum annealing (CSQA) using Hamiltonian

    .. math:: 

        H(s) = s H_p + \\Gamma (1-s)\\sum_i \\sigma_i^x

    where :math:`H_p` is the problem Hamiltonian we want to solve.

    Args:
        beta (float): Inverse temperature.
        gamma (float): Amplitude of quantum fluctuation.
        schedule (list): schedule list
        step_num (int): Number of Monte Carlo step.
        schedule_info (dict): Information about a annealing schedule.
        num_reads (int): Number of iterations.
        num_sweeps (int): number of sweeps
        schedule_info (dict): Information about a annealing schedule.

    """
    def __init__(self,
                 beta=5.0, gamma=1.0,
                 num_sweeps=1000, schedule=None,
                 num_reads=1):

        self._default_params = {
            'beta': beta,
            'gamma': gamma,
            'num_sweeps': num_sweeps,
            'schedule': schedule,
            'num_reads': num_reads
        }

        self._params = {
            'beta': beta,
            'gamma': gamma,
            'num_sweeps': num_sweeps,
            'schedule': schedule,
            'num_reads': num_reads
        }

    def _get_result(self, system, model):
        info = {}
        info['spin_config'] = system.spin_config

        state = cxxjij.result.get_solution(system)

        return state, info

    def sample_ising(self, h, J,
                     beta=None, gamma=None,
                     num_sweeps=None, schedule=None,
                     num_reads=None,
                     initial_state=None, updater=None,
                     reinitialize_state=True, seed=None):
        """Sampling from the Ising model.

        Args:
            h (dict): linear biases
            J (dict): quadratic biases
            beta (float, optional): inverse temperature
            gamma (float, optional): strength of transverse field
            num_sweeps (int, optional): number of sampling.
            schedule (list, optional): schedule list
            num_reads (int, optional): number of iterations
            initial_state (optional): initial state of spins
            updater (str, optional): updater algorithm
            reinitialize_state (bool, optional): Re-initilization at each sampling. Defaults to True.
            seed (int, optional): Sampling seed.
        
        Returns:
            :class:`openjij.sampler.response.Response`: results

        Examples:
            
            for Ising case::

                >>> h = {0: -1, 1: -1, 2: 1, 3: 1}
                >>> J = {(0, 1): -1, (3, 4): -1}
                >>> sampler = oj.CSQASampler()
                >>> res = sampler.sample_ising(h, J)

            for QUBO case::

                >>> Q = {(0, 0): -1, (1, 1): -1, (2, 2): 1, (3, 3): 1, (4, 4): 1, (0, 1): -1, (3, 4): 1}
                >>> sampler = oj.CSQASampler()
                >>> res = sampler.sample_qubo(Q)

        """

        #Set default updater
        if updater is None:
            updater='swendsenwang'

        bqm = openjij.BinaryQuadraticModel(
            linear=h, quadratic=J, vartype='SPIN', sparse=True
        )

        #Continuous time ising system only supports sparse ising graph

        ising_graph = bqm.get_cxxjij_ising_graph()

        self._set_params(
            beta=beta, gamma=gamma,
            num_sweeps=num_sweeps, num_reads=num_reads
        )
        self._annealing_schedule_setting(
            bqm, self._params['beta'], self._params['gamma'], self._params['num_sweeps'], self._params['schedule'])

        # make init state generator --------------------------------
        if initial_state is None:
            def init_generator():
                spin_config = np.random.choice([1,-1], len(bqm.variables))
                return list(spin_config)
        else:
            def init_generator(): return initial_state
        # -------------------------------- make init state generator

        # choose updater -------------------------------------------
        sqa_system = cxxjij.system.make_continuous_time_ising(
            init_generator(), ising_graph, self.gamma
        )
        _updater_name = updater.lower().replace('_', '').replace(' ', '')
        if _updater_name == 'swendsenwang':
            algorithm = cxxjij.algorithm.Algorithm_ContinuousTimeSwendsenWang_run
        else:
            raise ValueError('updater is one of "swendsen wang"')
        # ------------------------------------------- choose updater

        response = self._cxxjij_sampling(
            bqm, init_generator,
            algorithm, sqa_system,
            reinitialize_state, seed
        )

        response.info['schedule'] = self.schedule_info

        return response



import openjij as oj
import cxxjij as cj
import numpy as np

import unittest


class TestSamplers(unittest.TestCase):
    def setUp(self):
        self.num_ind = {
            'h': {0: -1, 1: -1, 2: 1, 3: 1},
            'J': {(0, 1): -1, (3, 4): -1}
        }
        str_ind = ['a', 'b', 'c', 'd', 'e']
        self.str_ising = {
            'h': {str_ind[i] for i in self.num_ind['h'].keys()},
            'J': {(str_ind[i], str_ind[j]) for i, j in self.num_ind['J'].keys()}
        }
        self.ground_state = [1, 1, -1, -1, -1]
        self.e_g = -1-1-1-1 + (-1-1)
        self.g_sample = {i: self.ground_state[i]
                         for i in range(len(self.ground_state))}
        self.g_samp_str = {k: self.ground_state[i]
                           for i, k in enumerate(str_ind)}

        self.qubo = {
            (0, 0): -1, (1, 1): -1, (2, 2): 1, (3, 3): 1, (4, 4): 1,
            (0, 1): -1, (3, 4): 1
        }
        self.str_qubo = {(str_ind[i], str_ind[j]): qij
                         for (i, j), qij in self.qubo.items()}
        # qubo (ndarray)
        self.qubo_ndarray = np.array(
                [[-1,-1, 0, 0, 0],
                 [ 0,-1, 0, 0, 0],
                 [ 0, 0, 1, 0, 0],
                 [ 0, 0, 0, 1, 1],
                 [ 0, 0, 0, 0, 1]])

        self.ground_q = [1, 1, 0, 0, 0]
        self.e_q = -1-1-1

        # for antiferromagnetic one-dimensional Ising model
        N = 30
        self.afih = {0: -10}
        self.afiJ = {(i, i+1): 1 for i in range(N-1)}
        self.afiground = {i:(-1)**i for i in range(N)}

    def samplers(self, sampler, init_state=None, init_q_state=None, schedule=None):
        res = sampler.sample_ising(
            self.num_ind['h'], self.num_ind['J'], schedule=schedule,
            initial_state=init_state, seed=1)
        self._test_response(res, self.e_g, self.ground_state)
        res = sampler.sample_qubo(self.qubo,
                                  initial_state=init_q_state, schedule=schedule, seed=2)
        self._test_response(res, self.e_q, self.ground_q)
        res = sampler.sample_qubo(self.qubo_ndarray,
                                  initial_state=init_q_state, schedule=schedule, seed=2)
        self._test_response(res, self.e_q, self.ground_q)

    def _test_response(self, res, e_g, s_g):
        # test openjij response interface
        self.assertEqual(len(res.states), 1)
        self.assertListEqual(s_g, list(res.states[0]))
        self.assertEqual(res.energies[0], e_g)
        # test dimod interface
        self.assertEqual(len(res.record.sample), 1)
        self.assertListEqual(s_g, list(res.record.sample[0]))
        self.assertEqual(res.record.energy[0], e_g)

    def _test_response_num(self, res, num_reads):
        # test openjij response interface
        self.assertEqual(len(res.states), num_reads)
        self.assertEqual(len(res.energies), num_reads)
        # test dimod interface
        self.assertEqual(len(res.record.sample), num_reads)
        self.assertEqual(len(res.record.energy), num_reads)

    def _test_num_reads(self, sampler_cls):
        num_reads = 10
        sampler = sampler_cls()
        res = sampler.sample_ising(
            self.num_ind['h'], self.num_ind['J'],
            num_reads=num_reads,
            seed=2
        )
        self._test_response_num(res, num_reads)

        sampler = sampler_cls(num_reads=num_reads)
        res = sampler.sample_ising(
            self.num_ind['h'], self.num_ind['J'],
        )
        self._test_response_num(res, num_reads)


    def test_sa(self):
        sampler = oj.SASampler()
        self.samplers(sampler)
        self.samplers(sampler, 
            init_state=[1 for _ in range(len(self.ground_state))],
            init_q_state=[1 for _ in range(len(self.ground_state))])
        self.samplers(sampler, 
            init_state={i: 1 for i in range(len(self.ground_state))}
            )

        # schedule [[beta, one_mc_steps], ...]
        # schedule test (list of list)
        self.samplers(sampler, 
            init_state={i: 1 for i in range(len(self.ground_state))},
            schedule=[[0.1, 10], [1, 10], [10, 10]]
            )

        # schedule test (list of tuple)
        self.samplers(sampler, 
            init_state={i: 1 for i in range(len(self.ground_state))},
            schedule=[(0.1, 10), (1, 10), (10, 10)]
            )

        self._test_num_reads(oj.SASampler)

        #antiferromagnetic one-dimensional Ising model
        sampler = oj.SASampler(num_reads=100)
        res = sampler.sample_ising(self.afih, self.afiJ, seed=1)
        self.assertDictEqual(self.afiground, res.first.sample)
        #antiferromagnetic one-dimensional Ising model
        sampler = oj.SASampler(num_reads=100)
        res = sampler.sample_ising(self.afih, self.afiJ, updater='swendsen wang', seed=1)
        self.assertDictEqual(self.afiground, res.first.sample)

    def test_sa_sparse(self):
        #sampler = oj.SASampler()
        #self.samplers(sampler)
        #self.samplers(sampler, 
        #    init_state=[1 for _ in range(len(self.ground_state))],
        #    init_q_state=[1 for _ in range(len(self.ground_state))])
        #self.samplers(sampler, 
        #    init_state={i: 1 for i in range(len(self.ground_state))}
        #    )

        ## schedule [[beta, one_mc_steps], ...]
        ## schedule test (list of list)
        #self.samplers(sampler, 
        #    init_state={i: 1 for i in range(len(self.ground_state))},
        #    schedule=[[0.1, 10], [1, 10], [10, 10]]
        #    )

        ## schedule test (list of tuple)
        #self.samplers(sampler, 
        #    init_state={i: 1 for i in range(len(self.ground_state))},
        #    schedule=[(0.1, 10), (1, 10), (10, 10)]
        #    )

        #self._test_num_reads(oj.SASampler)

        #antiferromagnetic one-dimensional Ising model
        sampler = oj.SASampler(num_reads=100)
        res = sampler.sample_ising(self.afih, self.afiJ, sparse=True, seed=1)
        self.assertDictEqual(self.afiground, res.first.sample)

    def test_sa_with_negative_interactions(self):
        # sa with negative interactions
        sampler = oj.SASampler()
        sampler.sample_ising({}, {(0,1): -1})
        sampler.sample_ising({2:-1}, {(0,1): -1})

    def test_sqa(self):
        sampler = oj.SQASampler()
        
        self.samplers(sampler)
        self.samplers(sampler, 
            init_state=[1 for _ in range(len(self.ground_state))],
            init_q_state=[1 for _ in range(len(self.ground_state))])
        self.samplers(sampler, 
            init_state={i: 1 for i in range(len(self.ground_state))}
            )

        # schedule [[s, one_mc_steps], ...]
        # schedule test (list of list, temperature fixed)
        self.samplers(sampler, 
            init_state={i: 1 for i in range(len(self.ground_state))},
            schedule=[[0.1, 10], [0.5, 10], [0.9, 10]]
            )

        # schedule test (list of tuple, temperature fixed)
        self.samplers(sampler, 
            init_state={i: 1 for i in range(len(self.ground_state))},
            schedule=[(0.1, 10), (0.5, 10), (0.9, 10)]
            )

        # schedule [[s, beta, one_mc_steps], ...]
        # schedule test (list of list, temperature non-fixed)
        self.samplers(sampler, 
            init_state={i: 1 for i in range(len(self.ground_state))},
            schedule=[[0.1, 0.1, 10], [0.5, 1, 10], [0.9, 10, 10]]
            )

        # schedule test (list of tuple, temperature non-fixed)
        self.samplers(sampler, 
            init_state={i: 1 for i in range(len(self.ground_state))},
            schedule=[(0.1, 0.1, 10), (0.5, 1, 10), (0.9, 10, 10)]
            )

        self._test_num_reads(oj.SQASampler)

        #antiferromagnetic one-dimensional Ising model
        sampler = oj.SQASampler(num_reads=100)
        res = sampler.sample_ising(self.afih, self.afiJ, seed=1)
        self.assertDictEqual(self.afiground, res.first.sample)

    def test_sqa_with_negative_interactions(self):
        # sa with negative interactions
        sampler = oj.SQASampler()
        sampler.sample_ising({}, {(0,1): -1})
        sampler.sample_ising({2:-1}, {(0,1): -1})

    # currently disabled
    #def test_csqa(self):
    #    #FIXME: This test is instable. Make sure if there is no bug in ContinuousIsing solver.
    #    #FIXME: Or is there some intristic reasons for this instability?
    #    #sampler = oj.CSQASampler(gamma=5, num_sweeps=500)
    #    #self.samplers(sampler,
    #    #        init_state=[1 for _ in range(len(self.ground_state))],
    #    #        init_q_state=[1 for _ in range(len(self.ground_state))])

    #    #antiferromagnetic one-dimensional Ising model
    #    sampler = oj.CSQASampler(num_reads=200)
    #    res = sampler.sample_ising(self.afih, self.afiJ, seed=1)
    #    self.assertDictEqual(self.afiground, res.first.sample)


if __name__ == '__main__':
    unittest.main()



import numpy as np

def qubo_to_ising(mat: np.ndarray):
    """inplace-convert numpy matrix from qubo to ising.

    Args:
        mat (np.ndarray): numpy matrix

    """
    mat /= 4
    for i in range(mat.shape[0]):
        mat[i, i] += np.sum(mat[i, :])
    

def chimera_to_ind(r: int, c: int, z: int, L: int):
    """[summary]

    Args:
        r (int): row index
        c (int): column index
        z (int): in-chimera index (must be from 0 to 7)
        L (int): height and width of  chimera-units (total number of spins is :math:`L \\times L \\times 8`)

    Raises:
        ValueError: [description]

    Returns:
        int: corresponding Chimera index
    """
    if not (0 <= r < L and 0 <= c < L and 0 <= z < 8):
        raise ValueError(
            '0 <= r < L or 0 <= c < L or 0 <= z < 8. '
            'your input r={}, c={}, z={}, L={}'.format(r, c, z, L))
    return r * L * 8 + c*8 + z



# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import openjij
from .model import make_BinaryQuadraticModel, make_BinaryQuadraticModel_from_JSON, BinaryQuadraticModel


def make_KingGraph(linear=None, quadratic=None, king_graph=None):
    """ KingGraph factory
    Returns:
        generated KingGraph class
    """
    mock_linear = {}
    mock_quadratic = {}

    if linear != None:
        mock_linear = linear

    if quadratic != None:
        mock_quadratic = quadratic

    if mock_linear == {} and mock_quadratic == {}:
        # no elements in linear and quadratic
        # fetch first element of the king_graph
        label = (king_graph[0][0], king_graph[0][0])
        # add to linear
        mock_linear[label] = 1.0

    class KingGraph(make_BinaryQuadraticModel(mock_linear, mock_quadratic)):
        """
        BQM for king graph of HITACHI CMOS Annealer
        Attributes:
            xrange (list(int)): represents hardware (CMOS) restricts for coordinate. [xmin, xmax]
            yrange (list(int)): represents hardware (CMOS) restricts for coordinate. [ymin, ymax]
            prange (list(int)): represents hardware (CMOS) restricts for the strength of interactions 'p'. [pmin, pmax]
            king_graph (list(int)): 
                Annealing cloud Web API format representation of interaction coefficients
                Quadratic term [x1, y1, x2, y2, value]
                Linear term    [x1, y1, x1, y1, value]
        """
    
        def __init__(self, linear: dict=None, quadratic: dict=None, offset: float=0.0,

                king_graph=None, vartype=openjij.SPIN, machine_type: str=''):
            """__init__.

            Args:
                linear (dict): linear biases
                quadratic (dict): quadratic biases
                offset (float): offset
                king_graph: represents ising or QUBO interaction.
                    Each spins are decided by 2-d corrdinate x, y.
                    Quadratic term [x1, y1, x2, y2, value]
                    Linear term    [x1, y1, x1, y1, value]
                vartype: 'SPIN' or 'BINARY'
                machine_type (str): choose 'ASIC' or 'FPGA'
            """
            vartype = openjij.cast_vartype(vartype)
    
            # set parameter ranges
            self.machine_type = machine_type
            if self.machine_type == "ASIC":
                self.xrange = [0, 351+1]
                self.yrange = [0, 175+1]
                self.prange = [-3, 3]
            elif self.machine_type == "FPGA":
                self.xrange = [0, 79+1]
                self.yrange = [0, 79+1]
                self.prange = [-127, 127]
            else:
                raise ValueError('machine type should be ASIC or FPGA')
    
            # convert format h, J, Q and initilize BQM
            if king_graph is not None:
                linear, quadratic = self._convert_to_BQM_format(
                    king_graph, vartype)
            super().__init__(linear, quadratic, offset=offset, vartype=vartype)
    
            # reformat to ising king graph (which is Web API format)
            if king_graph is not None and vartype == openjij.SPIN:
                self._ising_king_graph = king_graph
    
            else:
                # generate Ising h and J and create ising_king_graph format
                lin, quad, _ = self.to_ising()
                self._ising_king_graph = []
                for index, h in lin.items():
                    if h != 0:
                        x, y = self._convert_to_xy(index)
                        self._ising_king_graph.append([x, y, x, y, h])
                for (i, j), J in quad.items():
                    if J != 0:
                        x1, y1 = self._convert_to_xy(i)
                        x2, y2 = self._convert_to_xy(j)
                        self._ising_king_graph.append([x1, y1, x2, y2, J])
    
            self._validation_ising_king_graph()
    
        def _convert_to_BQM_format(self, king_graph, vartype):
            linear, quad = {}, {}
            for x1, y1, x2, y2, value in king_graph:
                if (x1, y1) == (x2, y2):
                    linear[(x1, y1)] = value
                else:
                    quad[(x1, y1), (x2, y2)] = value
            return linear, quad
    
        def get_ising_king_graph(self):
            return self._ising_king_graph
    
        def king_indices(self):
            if isinstance(self.indices[0], tuple):
                return self.indices
            else:
                return [self._convert_to_xy(i) for i in self.indices]
    
        def _convert_to_xy(self, index):
            if isinstance(index, tuple):
                return index[0], index[1]
            else:
                y = int(index / self.xrange[1])
                return int(index - y * self.xrange[1]), y
    
        def convert_to_index(self, x, y):
            return y * self.xrange[1] + x
    
        def _validation_ising_king_graph(self):
            for xi, yi, xj, yj, p in self._ising_king_graph:
                if yi >= self.yrange[1] or yj >= self.yrange[1]:
                    raise ValueError('Graph is incomplete xi: {}, yi: {}, xj: {}, yj: {}, p:{}'
                                     .format(xi, yi, xj, yj, p))
                if not (xi in [xj, xj-1, xj+1]) or not (yi in [yj, yj-1, yj+1]):
                    raise ValueError('Graph is incomplete xi: {}, yi: {}, xj: {}, yj: {}, p:{}'
                                     .format(xi, yi, xj, yj, p))
                if not (self.prange[0] <= p <= self.prange[1]):
                    raise ValueError('Graph is incomplete xi: {}, yi: {}, xj: {}, yj: {}, p: {}'
                                     .format(xi, yi, xj, yj, p))
    
    
        def energy(self, sample):
            return super().energy(sample, sparse=True)
    
        def energies(self, samples_like):
            return super().energies(sample_like, sparse=True)

    return KingGraph

def make_KingGraph_from_JSON(obj):
    """ KingGraph factory for JSON
    Args:
        obj (dict): JSON object
    Returns:
        generated KingGraph class
    """
    label = obj['variable_labels'][0]
    if isinstance(label, list):
        #convert to tuple
        label = tuple(label)

    mock_linear = {label:1.0}

    return make_KingGraph(mock_linear, {})

def KingGraph(linear=None, quadratic=None, offset=0.0,
                     king_graph=None, vartype=openjij.SPIN, machine_type=''):
    """generate KingGraph model.

    Args:
        linear (dict): linear biases
        quadratic (dict): quadratic biases
        offset (float): offset
        king_graph: represents ising or QUBO interaction.
            Each spins are decided by 2-d corrdinate x, y.
            
            * Quadratic term: [x1, y1, x2, y2, value]
            * Linear term:    [x1, y1, x1, y1, value]
        vartype: 'SPIN' or 'BINARY'
        machine_type (str): choose 'ASIC' or 'FPGA'
    Returns:
        generated KingGraphModel
    Examples:
        The following code shows intialization of KingGraph::

            >>> h = {}
            >>> J = {(0, 1): -1.0, (1, 2): -3.0}
            >>> king_graph = oj.KingGraph(machine_type="ASIC", linear=h, quadratic=J)

        You can initialize it from `king_interaction`::

            >>> king_interaction = [[0, 0, 1, 0, -1.0], [1, 0, 2, 0, -3.0]]
            >>> king_graph = oj.KingGraph(machine_type="ASIC", king_graph=king_interaction)
    """

    Model = make_KingGraph(linear, quadratic, king_graph)

    return Model(linear, quadratic, offset, king_graph, vartype, machine_type)

#classmethods
KingGraph.from_qubo = \
        lambda Q, offset=0.0, **kwargs: make_KingGraph({}, Q).from_qubo(Q, offset, **kwargs)

KingGraph.from_ising = \
        lambda linear, quadratic, offset=0.0, **kwargs: make_KingGraph(linear, quadratic).from_ising(linear, quadratic, offset, **kwargs)

KingGraph.from_serializable = \
        lambda obj: make_KingGraph_from_JSON(obj).from_serializable(obj)




# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
| This module defines the BinaryQuadraticModel with the Hamiltonian,

.. math:: 

    H = \\sum_{i\\neq j} J_{ij}\sigma_i \sigma_j + \\sum_{i} h_{i}\sigma_i,

| in an Ising form and

.. math:: 

    H = \\sum_{ij} Q_{ij}x_i x_j + \\sum_{i} H_{i}x_i,

| in a QUBO form.
| The methods and usage are basically the same as `dimod <https://github.com/dwavesystems/dimod>`_.

"""

import cxxjij
import cimod
import cxxcimod
import dimod


def make_BinaryQuadraticModel(linear: dict, quadratic: dict, sparse):
    """BinaryQuadraticModel factory.

    Returns:
        generated BinaryQuadraticModel class
    """

    class BinaryQuadraticModel(cimod.make_BinaryQuadraticModel(linear, quadratic, sparse)):
        """Represents Binary quadratic model. 
           Indices are listed in self.indices.
        Attributes:
            vartype (dimod.Vartype): variable type SPIN or BINARY
            linear (dict): represents linear term
            quadratic (dict): represents quadratic term
            offset (float): represents constant energy term when convert to SPIN from BINARY
            num_variables (int): represents number of variables in the model
            variables (list): represents variables of the binary quadratic model
        """

        def __init__(self, *args, **kwargs):
            """BinaryQuadraticModel constructor.

            Args:
                linear (dict): linear biases.
                quadratic (dict): quadratic biases
                offset (float): offset
                vartype (openjij.variable_type.Vartype): vartype
                gpu (bool): if true, this can be used for gpu samplers.
                kwargs:
            """
            # extract gpu keyword
            gpu = kwargs.pop('gpu', False)
            super().__init__(*args, **kwargs)
            self.gpu = gpu
            self.model_type = 'openjij.BinaryQuadraticModel'

        def get_cxxjij_ising_graph(self):
            """generate cxxjij Ising graph from the interactions.

            Returns:
                cxxjij.graph.Dense or cxxjij.graph.Sparse: 
                offset (float): offset of the energy due to qubo->ising transformation
            """

            if sparse:
                old_vartype = self.vartype
                self.change_vartype('SPIN')

                GraphClass = cxxjij.graph.Sparse if self.gpu == False else cxxjij.graph.SparseGPU
                offset = self.offset
                serialized = self.to_serializable()

                self.change_vartype(old_vartype)
                return GraphClass(serialized), offset
            else:
                old_vartype = self.vartype
                self.change_vartype('SPIN')

                GraphClass = cxxjij.graph.Dense if self.gpu == False else cxxjij.graph.DenseGPU
                # initialize with interaction matrix.
                mat = self.interaction_matrix()
                num_variables = mat.shape[0] - 1
                dense = GraphClass(num_variables)
                dense.set_interaction_matrix(mat)
                offset = self.offset

                self.change_vartype(old_vartype)
                return dense, offset

        # compatible with the previous version

        def calc_energy(self, sample, **kwargs):
            return self.energy(sample, **kwargs)

        # compatible with the previous version
        @property
        def indices(self):
            return self.variables

    return BinaryQuadraticModel


def make_BinaryQuadraticModel_from_JSON(obj: dict):
    """make BinaryQuadraticModel from JSON.

    Returns:
        corresponding BinaryQuadraticModel type
    """
    label = obj['variable_labels'][0]
    if isinstance(label, list):
        # convert to tuple
        label = tuple(label)

    mock_linear = {label: 1.0}

    if obj['version']['bqm_schema'] == '3.0.0-dense':
        sparse = False
    elif obj['version']['bqm_schema'] == '3.0.0':
        sparse = True
    else:
        raise TypeError("Invalid bqm_schema")

    return make_BinaryQuadraticModel(mock_linear, {}, sparse)


def BinaryQuadraticModel(linear, quadratic, *args, **kwargs):
    """generate BinaryQuadraticModel object.

    Attributes:
        vartype (dimod.Vartype): variable type SPIN or BINARY
        linear (dict): represents linear term
        quadratic (dict): represents quadratic term
        offset (float): represents constant energy term when convert to SPIN from BINARY
        num_variables (int): represents number of variables in the model
        variables (list): represents variables of the binary quadratic model
    Args:
        linear (dict): linear biases
        quadratic (dict): quadratic biases
        offset (float): offset
        vartype (openjij.variable_type.Vartype): vartype ('SPIN' or 'BINARY')
        gpu (bool): if true, this can be used for gpu samplers.
        kwargs:
    Returns:
        generated BinaryQuadraticModel
    Examples:
        BinaryQuadraticModel can be initialized by specifing h and J::

            >>> h = {0: 1, 1: -2}
            >>> J = {(0, 1): -1, (1, 2): -3, (2, 3): 0.5}
            >>> bqm = oj.BinaryQuadraticModel(self.h, self.J)

        You can also use strings and tuples of integers (up to 4 elements) as indices::

            >>> h = {'a': 1, 'b': -2}
            >>> J = {('a', 'b'): -1, ('b', 'c'): -3, ('c', 'd'): 0.5}
            >>> bqm = oj.BinaryQuadraticModel(self.h, self.J)
    """

    Model = make_BinaryQuadraticModel(
        linear, quadratic, kwargs.pop('sparse', False))

    def __extract_offset_and_vartype(*args, **kwargs):
        if kwargs == {}:
            if len(args) == 0:
                raise TypeError(
                    f"Offset or vartype is configured incorrectly. Vartype must be set.")
            elif len(args) == 1:
                offset = 0.0
                [vartype] = args
            elif len(args) == 2:
                [offset, vartype] = args
            else:
                raise TypeError(
                    f"Offset or vartype is configured incorrectly. Vartype must be set.")
        else:
            if 'offset' in kwargs and 'vartype' in kwargs:
                offset = kwargs['offset']
                vartype = kwargs['vartype']
            elif 'offset' in kwargs:
                if len(args) != 1:
                    raise TypeError(
                        f"Offset or vartype is configured incorrectly. Vartype must be set.")
                offset = kwargs['offset']
                [vartype] = args
            elif 'vartype' in kwargs:
                if len(args) >= 2:
                    raise TypeError(
                        f"Offset or vartype is configured incorrectly. Vartype must be set.")
                elif len(args) == 0:
                    offset = 0.0
                elif len(args) == 1:
                    [offset] = args
                vartype = kwargs['vartype']
            else:
                raise TypeError(
                    f"Offset or vartype is configured incorrectly. Vartype must be set.")

        return offset, vartype

    offset, vartype = __extract_offset_and_vartype(*args, **kwargs)

    return Model(linear, quadratic, offset, vartype)


# classmethods
def bqm_from_numpy_matrix(mat, variables: list = None, offset=0.0, vartype='BINARY', **kwargs):
    if variables is None:
        # generate array
        num_variables = mat.shape[0]
        variables = list(range(num_variables))

    return make_BinaryQuadraticModel({variables[0]: 1.0}, {}, kwargs.pop('sparse', False)).from_numpy_matrix(mat, variables, offset, vartype, True, **kwargs)


BinaryQuadraticModel.from_numpy_matrix = bqm_from_numpy_matrix

BinaryQuadraticModel.from_qubo = \
    lambda Q, offset=0.0, **kwargs: make_BinaryQuadraticModel(
        {}, Q, kwargs.pop('sparse', False)).from_qubo(Q, offset, **kwargs)

BinaryQuadraticModel.from_qubo = \
    lambda Q, offset=0.0, **kwargs: make_BinaryQuadraticModel(
        {}, Q, kwargs.pop('sparse', False)).from_qubo(Q, offset, **kwargs)

BinaryQuadraticModel.from_ising = \
    lambda linear, quadratic, offset=0.0, **kwargs: make_BinaryQuadraticModel(
        linear, quadratic, kwargs.pop('sparse', False)).from_ising(linear, quadratic, offset, **kwargs)

BinaryQuadraticModel.from_serializable = \
    lambda obj, **kwargs: make_BinaryQuadraticModel_from_JSON(
        obj).from_serializable(obj, **kwargs)


def make_BinaryPolynomialModel(polynomial, index_type=None, tuple_size=0):

    class BinaryPolynomialModel(cimod.make_BinaryPolynomialModel(polynomial, index_type, tuple_size)):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.model_type = 'openjij.BinaryPolynomialModel'

        def get_cxxjij_ising_graph(self):
            return cxxjij.graph.Polynomial(self.to_serializable())

        def calc_energy(self, sample, omp_flag=True):
            return self.energy(sample, omp_flag)

    return BinaryPolynomialModel


def make_BinaryPolynomialModel_from_JSON(obj):
    if(obj["type"] != "BinaryPolynomialModel"):
        raise Exception("Type must be \"BinaryPolynomialModel\"")
    mock_polynomial = {}
    if obj["index_type"] == 'IndexType.INT':
        mock_polynomial = {(0, 1): 1}
    elif obj["index_type"] == 'IndexType.STRING':
        mock_polynomial = {("a", "b"): 1}
    elif obj["index_type"] == 'IndexType.INT_TUPLE_2':
        mock_polynomial = {((0, 1), (1, 2)): 1}
    elif obj["index_type"] == 'IndexType.INT_TUPLE_3':
        mock_polynomial = {((0, 1, 2), (1, 2, 3)): 1}
    elif obj["index_type"] == 'IndexType.INT_TUPLE_4':
        mock_polynomial = {((0, 1, 2, 3), (1, 2, 3, 4)): 1}
    else:
        raise TypeError("Invalid types of polynomial")
    return make_BinaryPolynomialModel(mock_polynomial)


def BinaryPolynomialModel(*args, **kwargs):
    if kwargs == {}:
        if len(args) <= 1:
            raise TypeError("Invalid argument for this function")
        elif len(args) == 2:
            if isinstance(args[0], dict):
                return _BinaryPolynomialModel_from_dict(args[0], _to_cxxcimod(args[1]))
            else:
                raise TypeError("Invalid argument for this function")
        elif len(args) == 3:
            key_condition = isinstance(
                args[0], list) or isinstance(args[0], tuple)
            val_condition = isinstance(
                args[1], list) or isinstance(args[1], tuple)
            if key_condition and val_condition:
                return _BinaryPolynomialModel_from_list(args[0], args[1], _to_cxxcimod(args[2]))
            else:
                raise TypeError("Invalid argument for this function")
        else:
            raise TypeError("Invalid argument for this function")
    else:
        if 'keys' in kwargs and 'values' in kwargs and 'vartype' in kwargs:
            key_condition = isinstance(
                kwargs['keys'], list) or isinstance(kwargs['keys'], tuple)
            val_condition = isinstance(kwargs['values'], list) or isinstance(
                kwargs['values'], tuple)
            if key_condition and val_condition:
                return _BinaryPolynomialModel_from_list(kwargs['keys'], kwargs['values'], _to_cxxcimod(kwargs['vartype']))
            else:
                raise TypeError("Invalid argument for this function")
        elif 'polynomial' in kwargs and 'vartype' in kwargs:
            if isinstance(kwargs['polynomial'], dict):
                return _BinaryPolynomialModel_from_dict(kwargs['polynomial'], _to_cxxcimod(kwargs['vartype']))
            else:
                raise TypeError("Invalid argument for this function")
        elif 'values' in kwargs and 'vartype' in kwargs:
            if len(args) != 1:
                raise TypeError("Invalid argument for this function")
            key_condition = isinstance(
                args[0], list) or isinstance(args[0], tuple)
            val_condition = isinstance(kwargs['values'], list) or isinstance(
                kwargs['values'], tuple)
            if key_condition and val_condition:
                return _BinaryPolynomialModel_from_list(args[0], kwargs['values'], _to_cxxcimod(kwargs['vartype']))
            else:
                raise TypeError("Invalid argument for this function")
        elif 'vartype' in kwargs:
            if len(args) == 1:
                if isinstance(args[0], dict):
                    return _BinaryPolynomialModel_from_dict(args[0], _to_cxxcimod(kwargs['vartype']))
                else:
                    raise TypeError("Invalid argument for this function")
            elif len(args) == 2:
                key_condition = isinstance(
                    args[0], list) or isinstance(args[0], tuple)
                val_condition = isinstance(
                    args[1], list) or isinstance(args[1], tuple)
                if key_condition and val_condition:
                    return _BinaryPolynomialModel_from_list(args[0], args[1], _to_cxxcimod(kwargs['vartype']))
                else:
                    raise TypeError("Invalid argument for this function")
            else:
                raise TypeError("Invalid argument for this function")
        else:
            raise TypeError("Invalid argument for this function")


def _BinaryPolynomialModel_from_dict(polynomial: dict, vartype):
    Model = make_BinaryPolynomialModel(polynomial)
    return Model(polynomial, _to_cxxcimod(vartype))


def _BinaryPolynomialModel_from_list(keys: list, values: list, vartype):
    if len(keys) == 0:
        Model = make_BinaryPolynomialModel({})
        return Model(keys, values, _to_cxxcimod(vartype))
    i = 0
    label = None
    while i < len(keys):
        if len(keys[i]) > 0:
            label = keys[i][0]
            break
        i += 1
    if label == None:
        Model = make_BinaryPolynomialModel({(): 1.0})
        return Model(keys, values, _to_cxxcimod(vartype))
    else:
        if isinstance(label, list):
            label = tuple(label)
        mock_polynomial = {(label,): 1.0}
        Model = make_BinaryPolynomialModel(mock_polynomial)
        return Model(keys, values, _to_cxxcimod(vartype))


def make_BinaryPolynomialModel_from_hising(*args, **kwargs):
    if kwargs == {}:
        if len(args) == 0:
            raise TypeError("Invalid argument for this function")
        elif len(args) == 1:
            if isinstance(args[0], dict):
                return _make_BinaryPolynomialModel_from_hising_from_dict(args[0])
            else:
                raise TypeError("Invalid argument for this function")
        elif len(args) == 2:
            key_condition = isinstance(
                args[0], list) or isinstance(args[0], tuple)
            val_condition = isinstance(
                args[1], list) or isinstance(args[1], tuple)
            if key_condition and val_condition:
                return _make_BinaryPolynomialModel_from_hising_from_list(args[0], args[1])
            else:
                raise TypeError("Invalid argument for this function")
        else:
            raise TypeError("Invalid argument for this function")
    else:
        if 'keys' in kwargs and 'values' in kwargs:
            key_condition = isinstance(
                kwargs['keys'], list) or isinstance(kwargs['keys'], tuple)
            val_condition = isinstance(kwargs['values'], list) or isinstance(
                kwargs['values'], tuple)
            if key_condition and val_condition:
                return _make_BinaryPolynomialModel_from_hising_from_list(kwargs['keys'], kwargs['values'])
            else:
                raise TypeError("Invalid argument for this function")
        elif 'values' in kwargs:
            if len(args) != 1:
                raise TypeError("Invalid argument for this function")
            key_condition = isinstance(
                args[0], list) or isinstance(args[0], tuple)
            val_condition = isinstance(kwargs['values'], list) or isinstance(
                kwargs['values'], tuple)
            if key_condition and val_condition:
                return _make_BinaryPolynomialModel_from_hising_from_list(args[0], kwargs['values'])
            else:
                raise TypeError("Invalid argument for this function")
        elif 'polynomial' in kwargs:
            if len(args) != 0:
                raise TypeError("Invalid argument for this function")
            if isinstance(kwargs['polynomial'], dict):
                _make_BinaryPolynomialModel_from_hising_from_dict(
                    kwargs['polynomial'])
            else:
                raise TypeError("Invalid argument for this function")
        else:
            raise TypeError("Invalid argument for this function")


def _make_BinaryPolynomialModel_from_hising_from_dict(polynomial: dict):
    return make_BinaryPolynomialModel(polynomial).from_hising(polynomial)


def _make_BinaryPolynomialModel_from_hising_from_list(keys: list, values: list):
    if len(keys) == 0:
        return make_BinaryPolynomialModel({}).from_hising(keys, values)

    i = 0
    label = None
    while i < len(keys):
        if len(keys[i]) > 0:
            label = keys[i][0]
            break
        i += 1

    if label == None:
        return make_BinaryPolynomialModel({(): 1.0}).from_hising(keys, values)
    else:
        if isinstance(label, list):
            label = tuple(label)
        mock_polynomial = {(label,): 1.0}
        return make_BinaryPolynomialModel(mock_polynomial).from_hising(keys, values)


def make_BinaryPolynomialModel_from_hubo(*args, **kwargs):
    if kwargs == {}:
        if len(args) == 0:
            raise TypeError("Invalid argument for this function")
        elif len(args) == 1:
            if isinstance(args[0], dict):
                return _make_BinaryPolynomialModel_from_hubo_from_dict(args[0])
            else:
                raise TypeError("Invalid argument for this function")
        elif len(args) == 2:
            key_condition = isinstance(
                args[0], list) or isinstance(args[0], tuple)
            val_condition = isinstance(
                args[1], list) or isinstance(args[1], tuple)
            if key_condition and val_condition:
                return _make_BinaryPolynomialModel_from_hubo_from_list(args[0], args[1])
            else:
                raise TypeError("Invalid argument for this function")
        else:
            raise TypeError("Invalid argument for this function")
    else:
        if 'keys' in kwargs and 'values' in kwargs:
            key_condition = isinstance(
                kwargs['keys'], list) or isinstance(kwargs['keys'], tuple)
            val_condition = isinstance(kwargs['values'], list) or isinstance(
                kwargs['values'], tuple)
            if key_condition and val_condition:
                return _make_BinaryPolynomialModel_from_hubo_from_list(kwargs['keys'], kwargs['values'])
            else:
                raise TypeError("Invalid argument for this function")
        elif 'values' in kwargs:
            if len(args) != 1:
                raise TypeError("Invalid argument for this function")
            key_condition = isinstance(
                args[0], list) or isinstance(args[0], tuple)
            val_condition = isinstance(kwargs['values'], list) or isinstance(
                kwargs['values'], tuple)
            if key_condition and val_condition:
                return _make_BinaryPolynomialModel_from_hubo_from_list(args[0], kwargs['values'])
            else:
                raise TypeError("Invalid argument for this function")
        elif 'polynomial' in kwargs:
            if len(args) != 0:
                raise TypeError("Invalid argument for this function")
            if isinstance(kwargs['polynomial'], dict):
                _make_BinaryPolynomialModel_from_hubo_from_dict(
                    kwargs['polynomial'])
            else:
                raise TypeError("Invalid argument for this function")
        else:
            raise TypeError("Invalid argument for this function")


def _make_BinaryPolynomialModel_from_hubo_from_dict(polynomial: dict):
    return make_BinaryPolynomialModel(polynomial).from_hubo(polynomial)


def _make_BinaryPolynomialModel_from_hubo_from_list(keys: list, values: list):
    if len(keys) == 0:
        return make_BinaryPolynomialModel({}).from_hubo(keys, values)

    i = 0
    label = None
    while i < len(keys):
        if len(keys[i]) > 0:
            label = keys[i][0]
            break
        i += 1

    if label == None:
        return make_BinaryPolynomialModel({(): 1.0}).from_hubo(keys, values)
    else:
        if isinstance(label, list):
            label = tuple(label)
        mock_polynomial = {(label,): 1.0}
        return make_BinaryPolynomialModel(mock_polynomial).from_hubo(keys, values)


def _to_cxxcimod(vartype):
    # convert to cxxcimod type
    if isinstance(vartype, cxxcimod.Vartype):
        return vartype

    vartype = dimod.as_vartype(vartype)
    if vartype == dimod.SPIN:
        return cxxcimod.Vartype.SPIN
    if vartype == dimod.BINARY:
        return cxxcimod.Vartype.BINARY


# classmethods
BinaryPolynomialModel.from_serializable = lambda obj: make_BinaryPolynomialModel_from_JSON(
    obj).from_serializable(obj)
BinaryPolynomialModel.from_hising = lambda *args, **kwargs: make_BinaryPolynomialModel_from_hising(
    *args, **kwargs)
BinaryPolynomialModel.from_hubo = lambda *args, **kwargs: make_BinaryPolynomialModel_from_hubo(
    *args, **kwargs)



# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
| OpenJij response class.
| The usage is almost the same as :class:`dimod.SampleSet`.
"""

import numpy as np
import dimod
import openjij


class Response(dimod.SampleSet):
    @property
    def states(self):
        return self.record.sample

    @property
    def indices(self):
        return list(self.variables)

    @property
    def energies(self):
        return self.record.energy

    @property
    def min_samples(self):
        return self.first.sample




# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import time
import inspect
from logging import getLogger
import scipy as sp


logger = getLogger(__name__)


def solver_benchmark(solver, time_list, solutions=[], args={}, p_r=0.99, ref_energy=0, measure_with_energy=False, time_name='execution_time'):
    """Calculate 'success probability', 'TTS', 'Residual energy','Standard Error' with computation time
    Args:
        solver (callable): returns openjij.Response, and solver has arguments 'time' and '**args'
        time_list (list):
        solutions (list(list(int)), list(int)): true solution or list of solution (if solutions are degenerated).
        args (dict): Arguments for solver.
        p_r (float): Thereshold probability for time to solutions.
        ref_energy (float): The ground (reference to calculate success probability and the residual energy) energy.
        measure_with_energy (bool): use a energy as measure for success
    Returns:
        dict: dictionary which has the following keys:

        * **time**: list of compuation time
        * **success_prob** list of success probability at each computation time
        * **tts**: list of time to solusion at each computation time
        * **residual_energy**: list of residual energy at each computation time
        * **se_lower_tts**: list of tts's lower standard error at each computation time
        * **se_upper_tts**: list of tts's upper standard error at each computation time
        * **se_success_prob**: list of success probability's standard error at each computation time
        * **se_residual_energy**: list of residual_energy's standard error at each computation time
        * **info** (dict): Parameter information for the benchmark
            
    """

    if not measure_with_energy:
        if solutions == []:
            raise ValueError("need input 'solutions': (list(list))")

    logger.info('function ' + inspect.currentframe().f_code.co_name + ' start')

    computation_times = []
    success_probabilities = []
    tts_list = []
    residual_energies = []

    se_lower_tts_list = []
    se_upper_tts_list = []
    se_success_prob_list = []
    se_residual_energy_list = []

    for time in time_list:
        response = solver(time, **args)

        comp_time = response.info[time_name]
        computation_times.append(comp_time)

        ps = success_probability(
            response, solutions, ref_energy, measure_with_energy)
        tts = time_to_solution(ps, comp_time, p_r)

        success_probabilities.append(ps)
        tts_list.append(tts)
        residual_energies.append(residual_energy(response, ref_energy))

        se_ps = se_success_probability(
            response, solutions, ref_energy, measure_with_energy)

        se_success_prob_list.append(se_ps)
        se_lower_tts_list.append(se_lower_tts(tts, ps, comp_time, p_r, se_ps))
        se_upper_tts_list.append(se_upper_tts(tts, ps, comp_time, p_r, se_ps))
        se_residual_energy_list.append(
            se_residual_energy(response, ref_energy))

    return {
        "time": computation_times,
        "success_prob": success_probabilities,
        "tts": tts_list,
        "residual_energy": residual_energies,
        "se_lower_tts": se_lower_tts_list,
        "se_upper_tts": se_upper_tts_list,
        "se_success_prob": se_success_prob_list,
        "se_residual_energy": se_residual_energy_list,
        "info": {
            "tts_threshold_prob": p_r,
            "ref_energy": ref_energy,
            "measure_with_energy": measure_with_energy
        }
    }


def residual_energy(response, ref_energy):
    """Calculate redisual energy from measure energy
    Args:
        response (openjij.Response): response from solver (or sampler).
        ref_energy (float): the reference energy (usually use the ground energy)
    Returns:
        float: Residual energy which is defined as :math:`\\langle E \\rangle - E_0` (:math:`\\langle...\\rangle` represents average, :math:`E_0` is the reference energy (usually use the ground energy)).

    """
    return np.mean(response.energies) - ref_energy


def se_residual_energy(response, ref_energy):
    """Calculate redisual energy's standard error from measure energy
    Args:
        response (openjij.Response): response from solver (or sampler).
        ref_energy (float): the reference energy (usually use the ground energy)
    Returns:
        float: redisual energy's standard error from measure energy
    """
    return np.std(response.energies, ddof=1)


def success_probability(response, solutions, ref_energy=0, measure_with_energy=False):
    """Calculate success probability from openjij.response
    Args:
        response (openjij.Response): response from solver (or sampler).
        solutions (list[int]): true solutions.
    Returns:
        float: Success probability.

        * When measure_with_energy is False, success is defined as getting the same state as solutions.
        * When measure_with_energy is True, success is defined as getting a state which energy is below reference energy

    """

    if measure_with_energy:
        suc_prob = np.count_nonzero(
            np.array(response.energies) <= ref_energy)/len(response.energies)
    else:
        if isinstance(solutions[0], dict):
            sampled_states = response.samples()
            suc_prob = np.mean([
                1 if dict(state) in solutions else 0
                for state in sampled_states
            ])
        else:
            sampled_states = response.states
            suc_prob = np.mean([
                1 if list(state) in solutions else 0
                for state in sampled_states
            ])

    return suc_prob


def se_success_probability(response, solutions, ref_energy=0, measure_with_energy=False):
    """Calculate success probability's standard error from openjij.response
    Args:
        response (openjij.Response): response from solver (or sampler).
        solutions (list[int]): true solutions.
    Returns:
        float: Success probability's standard error.

        * When measure_with_energy is False, success is defined as getting the same state as solutions.
        * When measure_with_energy is True, success is defined as getting a state which energy is below reference energy

    """

    if measure_with_energy:
        se_suc_prob = np.sqrt(np.count_nonzero(
            np.array(response.energies) <= ref_energy)/(len(response.energies)-1))
    else:
        if isinstance(solutions[0], dict):
            sampled_states = response.samples()
            se_suc_prob = np.std([
                1 if dict(state) in solutions else 0
                for state in sampled_states
            ])
        else:
            sampled_states = response.states
            se_suc_prob = np.std([
                1 if list(state) in solutions else 0
                for state in sampled_states
            ])

    return se_suc_prob


def time_to_solution(success_prob, computation_time, p_r):
    """
    Args:
        success_prob (float): success probability.
        computation_time (float):
        p_r (float): thereshold probability to calculate time to solution.
    Returns:
        float: time to solution :math:`\\tau * \\log(1-pr)/\\log(1-ps)` which pr is thereshold probability, ps is success probability and :math:`tau` is computation time.
    """

    if success_prob == 1.0:
        tts = 0.0
    elif success_prob == 0.0:
        tts = np.inf
    else:
        tts = computation_time * np.log(1 - p_r) / np.log(1-success_prob)

    return tts


def se_lower_tts(tts, success_prob, computation_time, p_r, se_success_prob):
    """
    Args:
        success_prob (float): success probability.
        computation_time (float):
        p_r (float): thereshold probability to calculate time to solution.
    Returns:
        float: time to solution :math:`\\tau * \\log(1-pr)/\\log(1-ps)` 's standard error which pr is thereshold probability, ps is success probability and :math:`tau` is computation time.
    """

    if 1 - (success_prob + se_success_prob) <= 0.0:
        tts_low_error = 0.0
    elif success_prob == 0.0:
        tts_low_error = 0.0
    else:
        tts_low_error = computation_time * \
            np.log(1 - p_r) / np.log(1 - (success_prob + se_success_prob))

    se_lower_tts = abs(tts_low_error - tts)

    return se_lower_tts


def se_upper_tts(tts, success_prob, computation_time, p_r, se_success_prob):
    """
    Args:
        success_prob (float): success probability.
        computation_time (float):
        p_r (float): thereshold probability to calculate time to solution.
    Returens:
        float: time to solution :math:`\\tau * \\log(1-pr)/\\log(1-ps)` 's standard error which pr is thereshold probability, ps is success probability and :math:`tau` is computation time.
    """

    if success_prob == 1.0:
        tts_up_error = 0.0
    elif success_prob == 0.0:
        tts_up_error = 0.0
    else:
        tts_up_error = computation_time * \
            np.log(1 - p_r) / np.log(1 - (success_prob - se_success_prob))

    se_upper_tts = abs(tts_up_error - tts)

    return se_upper_tts



# Copyright 2021 Jij Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License


import numpy as np
import openjij
from openjij.sampler import BaseSampler
from openjij.utils.graph_utils import qubo_to_ising
import cxxjij
import dimod
import cimod
from typing import Union, Optional

"""
This module contains Simulated Annealing sampler.
"""


class SASampler(BaseSampler):
    """Sampler with Simulated Annealing (SA).

    Args:
        beta_min (float):
            Minmum beta (inverse temperature).
            You can overwrite in methods .sample_*.

        beta_max (float):
            Maximum beta (inverse temperature).
            You can overwrite in methods .sample_*.

        num_reads (int):
            number of sampling (algorithm) runs. defaults None.
            You can overwrite in methods .sample_*.

        num_sweeps (int):
            number of MonteCarlo steps during SA. defaults None.
            You can overwrite in methods .sample_*.

        schedule_info (dict):
            Information about an annealing schedule.

    Raises:
        ValueError: If schedules or variables violate as below.
        - not list or numpy.array.
        - not list of tuple (beta : float, step_length : int).
        - beta is less than zero.
    """

    @property
    def parameters(self):
        return {
            'beta_min': ['parameters'],
            'beta_max': ['parameters'],
        }

    def __init__(self,
                 beta_min: Optional[float] = None,
                 beta_max: Optional[float] = None,
                 num_sweeps: Optional[int] = None,
                 num_reads: Optional[int] = None,
                 schedule: Optional[list] = None):

        # Set default parameters
        if num_sweeps is None:
            num_sweeps = 1000
        if num_reads is None:
            num_reads = 1

        self._default_params = {
            'beta_min': beta_min,
            'beta_max': beta_max,
            'num_sweeps': num_sweeps,
            'schedule': schedule,
            'num_reads': num_reads
        }

        self._params = self._default_params.copy()

        self._make_system = {
            'singlespinflip': cxxjij.system.make_classical_ising,
            'singlespinflippolynomial': cxxjij.system.make_classical_ising_polynomial,
            'swendsenwang': cxxjij.system.make_classical_ising
        }
        self._algorithm = {
            'singlespinflip': cxxjij.algorithm.Algorithm_SingleSpinFlip_run,
            'singlespinflippolynomial': cxxjij.algorithm.Algorithm_SingleSpinFlip_run,
            'swendsenwang': cxxjij.algorithm.Algorithm_SwendsenWang_run
        }

    def _convert_validation_schedule(self, schedule):
        """Checks if the schedule is valid and returns cxxjij schedule
        """
        if not isinstance(schedule, (list, np.array)):
            raise ValueError("schedule should be list or numpy.array")

        if isinstance(schedule[0], cxxjij.utility.ClassicalSchedule):
            return schedule

        if len(schedule[0]) != 2:
            raise ValueError(
                "schedule is list of tuple or list (beta : float, step_length : int)")

        # schedule validation  0 <= beta
        beta = np.array(schedule).T[0]
        if not np.all(0 <= beta):
            raise ValueError("schedule beta range is '0 <= beta'.")

        # convert to cxxjij.utility.ClassicalSchedule
        cxxjij_schedule = []
        for beta, step_length in schedule:
            _schedule = cxxjij.utility.ClassicalSchedule()
            _schedule.one_mc_step = step_length
            _schedule.updater_parameter.beta = beta
            cxxjij_schedule.append(_schedule)

        return cxxjij_schedule

    def sample(self,
               bqm: Union['openjij.BinaryQuadraticModel',
                          dimod.BinaryQuadraticModel],
               beta_min: Optional[float] = None,
               beta_max: Optional[float] = None,
               num_sweeps: Optional[int] = None,
               num_reads: Optional[int] = None,
               schedule: Optional[list] = None,
               initial_state: Optional[Union[list, dict]] = None,
               updater: Optional[str] = None,
               sparse: Optional[bool] = None,
               reinitialize_state: Optional[bool] = None,
               seed: Optional[int] = None) -> 'openjij.sampler.response.Response':

        """sample Ising model.

        Args:
            bqm (oj.BinaryQuadraticModel) binary quadratic model
            beta_min (float): minimal value of inverse temperature
            beta_max (float): maximum value of inverse temperature
            num_sweeps (int): number of sweeps
            num_reads (int): number of reads
            schedule (list): list of inverse temperature
            initial_state (dict): initial state
            updater(str): updater algorithm
            reinitialize_state (bool): if true reinitialize state for each run
            seed (int): seed for Monte Carlo algorithm
        Returns:
            :class:`openjij.sampler.response.Response`: results
            
        Examples:
            
            for Ising case::

                >>> h = {0: -1, 1: -1, 2: 1, 3: 1}
                >>> J = {(0, 1): -1, (3, 4): -1}
                >>> sampler = oj.SASampler()
                >>> res = sampler.sample_ising(h, J)

            for QUBO case::

                >>> Q = {(0, 0): -1, (1, 1): -1, (2, 2): 1, (3, 3): 1, (4, 4): 1, (0, 1): -1, (3, 4): 1}
                >>> sampler = oj.SASampler()
                >>> res = sampler.sample_qubo(Q)
            
        """

        # Set default parameters
        if updater is None:
            updater = 'single spin flip'
        if sparse is None:
            sparse = False
        if reinitialize_state is None:
            reinitialize_state = True

        _updater_name = updater.lower().replace('_', '').replace(' ', '')
        # swendsen wang algorithm runs only on sparse ising graphs.
        if _updater_name == 'swendsenwang' or sparse:
            sparse = True
        else:
            sparse = False

        if type(bqm) == dimod.BinaryQuadraticModel:
            bqm = openjij.model.BinaryQuadraticModel(dict(bqm.linear), dict(bqm.quadratic), bqm.offset, bqm.vartype, sparse=sparse)

        if sparse and bqm.sparse == False:
            # convert to sparse bqm
            bqm = openjij.model.BinaryQuadraticModel(bqm.linear, bqm.quadratic, bqm.offset, bqm.vartype, sparse=True)

        # alias 
        model = bqm

        ising_graph, offset = model.get_cxxjij_ising_graph()

        self._set_params(
            beta_min=beta_min, beta_max=beta_max,
            num_sweeps=num_sweeps, num_reads=num_reads,
            schedule=schedule
            )

        # set annealing schedule -------------------------------
        if self._params['schedule'] is None:
            self._params['schedule'], beta_range = geometric_ising_beta_schedule(
                model=model,
                beta_max=self._params['beta_max'],
                beta_min=self._params['beta_min'],
                num_sweeps=self._params['num_sweeps']
            )
            self.schedule_info = {
                'beta_max': beta_range[0],
                'beta_min': beta_range[1],
                'num_sweeps': self._params['num_sweeps']
            }
        else:
            self._params['schedule'] = self._convert_validation_schedule(self._params['schedule'])
            self.schedule_info = {'schedule': 'custom schedule'}
        # ------------------------------- set annealing schedule

        # make init state generator --------------------------------
        if initial_state is None:
            def _generate_init_state(): return ising_graph.gen_spin(seed) if seed != None else ising_graph.gen_spin()
        else:
            temp_initial_state = []
            if isinstance(initial_state, dict):
                if model.vartype == openjij.BINARY:
                    for k in model.variables:
                        v = initial_state[k]
                        if v != 0 and v != 1:
                            raise RuntimeError("The initial variables must be 0 or 1.")
                        temp_initial_state.append(2*v - 1)
                elif model.vartype == openjij.SPIN:
                    for k in model.variables:
                        v = initial_state[k]
                        if v != -1 and v != 1:
                            raise RuntimeError("The initial variables must be -1 or +1.")
                        temp_initial_state.append(v)
                else:
                    raise RuntimeError("Unknown vartype detected.")
            elif isinstance(initial_state, (list, tuple)):
                if model.vartype == openjij.BINARY:
                    for k in range(len(model.variables)):
                        v = initial_state[k]
                        if v != 0 and v != 1:
                            raise RuntimeError("The initial variables must be 0 or 1.")
                        temp_initial_state.append(2*v - 1)
                elif model.vartype == openjij.SPIN:
                    for k in range(len(model.variables)):
                        v = initial_state[k]
                        if v != -1 and v != 1:
                            raise RuntimeError("The initial variables must be -1 or +1.")
                        temp_initial_state.append(v)
                else:
                    raise RuntimeError("Unknown vartype detected.")
            else:
                raise RuntimeError("Unsupported type of initial_state.")

            _init_state = np.array(temp_initial_state)

            # validate initial_state size
            if len(initial_state) != ising_graph.size():
                raise ValueError(
                    "the size of the initial state should be {}"
                    .format(ising_graph.size()))

            def _generate_init_state(): return np.array(_init_state)
        # -------------------------------- make init state generator

        # choose updater -------------------------------------------
        _updater_name = updater.lower().replace('_', '').replace(' ', '')
        if _updater_name not in self._make_system:
            raise ValueError('updater is one of "single spin flip or swendsen wang"')
        algorithm = self._algorithm[_updater_name]
        sa_system = self._make_system[_updater_name](_generate_init_state(), ising_graph)
        # ------------------------------------------- choose updater
        response = self._cxxjij_sampling(
            model, _generate_init_state,
            algorithm, sa_system,
            reinitialize_state, seed
        )

        response.info['schedule'] = self.schedule_info

        return response

    def sample_hubo(self,
                    J: Union[dict,
                             'openjij.BinaryPolynomialModel',
                             cimod.BinaryPolynomialModel],
                    vartype: Optional[str] = None,
                    beta_min: Optional[float] = None,
                    beta_max: Optional[float] = None,
                    num_sweeps: Optional[int] = None,
                    num_reads: Optional[int] = None,
                    schedule: Optional[list] = None,
                    initial_state: Optional[Union[list, dict]] = None,
                    updater: Optional[str] = None,
                    reinitialize_state: Optional[bool] = None,
                    seed: Optional[int] = None) -> 'openjij.sampler.response.Response':

        """sampling from higher order unconstrainted binary optimization.

        Args:
            J (dict): Interactions.
            vartype (str, openjij.VarType): "SPIN" or "BINARY".
            beta_min (float, optional): Minimum beta (initial inverse temperature). Defaults to None.
            beta_max (float, optional): Maximum beta (final inverse temperature). Defaults to None.
            schedule (list, optional): schedule list. Defaults to None.
            num_sweeps (int, optional): number of sweeps. Defaults to None.
            num_reads (int, optional): number of reads. Defaults to 1.
            init_state (list, optional): initial state. Defaults to None.
            reinitialize_state (bool): if true reinitialize state for each run
            seed (int, optional): seed for Monte Carlo algorithm. Defaults to None.

        Returns:
            :class:`openjij.sampler.response.Response`: results

        Examples::
            for Ising case::
                >>> sampler = oj.SASampler()
                >>> J = {(0,): -1, (0, 1): -1, (0, 1, 2): 1}
                >>> response = sampler.sample_hubo(J, "SPIN")

            for Binary case::
                >>> sampler = oj.SASampler()
                >>> J = {(0,): -1, (0, 1): -1, (0, 1, 2): 1}
                >>> response = sampler.sample_hubo(J, "BINARY")
        """

        # Set default parameters
        if reinitialize_state is None:
            reinitialize_state = True

        # Set model
        if str(type(J)) == str(type(openjij.model.BinaryPolynomialModel({}, "SPIN"))):
            if vartype is not None:
                raise ValueError("vartype must not be specified")
            model = J
        elif str(type(J)) == str(type(cimod.BinaryPolynomialModel({}, "SPIN"))):
            if vartype is not None:
                raise ValueError("vartype must not be specified")
            model = J
        else:
            model = openjij.model.BinaryPolynomialModel(J, vartype)
  
        # make init state generator --------------------------------
        if initial_state is None:
            if model.vartype == openjij.SPIN:
                def _generate_init_state(): return cxxjij.graph.Polynomial(model.num_variables).gen_spin(seed) if seed != None else cxxjij.graph.Polynomial(model.num_variables).gen_spin()
            elif model.vartype == openjij.BINARY:
                def _generate_init_state(): return cxxjij.graph.Polynomial(model.num_variables).gen_binary(seed) if seed != None else cxxjij.graph.Polynomial(model.num_variables).gen_binary()
            else:
                raise ValueError("Unknown vartype detected")
        else:
            if isinstance(initial_state, dict):
                initial_state = [initial_state[k] for k in model.indices]
            def _generate_init_state(): return np.array(initial_state)
        # -------------------------------- make init state generator

        # determine system class and algorithm --------------------------------
        if model.vartype == openjij.SPIN:
            if updater is None or updater == "single spin flip":
                sa_system = cxxjij.system.make_classical_ising_polynomial(_generate_init_state(), model.to_serializable())
                algorithm = cxxjij.algorithm.Algorithm_SingleSpinFlip_run
            elif updater == "k-local":
                raise ValueError("k-local update is only supported for binary variables")
            else:
                raise ValueError("Unknown updater name")
        elif model.vartype == openjij.BINARY:
            if updater == "k-local":
                sa_system = cxxjij.system.make_k_local_polynomial(_generate_init_state(), model.to_serializable())
                algorithm = cxxjij.algorithm.Algorithm_KLocal_run
            elif updater is None or updater == "single spin flip":
                sa_system = cxxjij.system.make_classical_ising_polynomial(_generate_init_state(), model.to_serializable())
                algorithm = cxxjij.algorithm.Algorithm_SingleSpinFlip_run
            else:
                raise ValueError("Unknown updater name")
        else:
            raise ValueError("Unknown vartype detected")
        # -------------------------------- determine system class and algorithm

        self._set_params(
            beta_min=beta_min, beta_max=beta_max,
            num_sweeps=num_sweeps, num_reads=num_reads,
            schedule=schedule
            )

        # set annealing schedule -------------------------------
        if self._params['schedule'] is None:
            self._params['schedule'], beta_range = geometric_hubo_beta_schedule(
                sa_system, self._params['beta_max'],
                self._params['beta_min'], self._params['num_sweeps']
            )
            self.schedule_info = {
                'beta_max': beta_range[0],
                'beta_min': beta_range[1],
                'num_sweeps': self._params['num_sweeps']
            }
        else:
            self.schedule_info = {'schedule': 'custom schedule'}
        # ------------------------------- set annealing schedule

        response = self._cxxjij_sampling(
            model, _generate_init_state,
            algorithm, sa_system,
            reinitialize_state, seed
        )

        response.info['schedule'] = self.schedule_info

        return response

def geometric_ising_beta_schedule(model: openjij.model.BinaryQuadraticModel,
                                  beta_max=None, beta_min=None,
                                  num_sweeps=1000):
    """make geometric cooling beta schedule

    Args:
        model (openjij.model.BinaryQuadraticModel)
        beta_max (float, optional): [description]. Defaults to None.
        beta_min (float, optional): [description]. Defaults to None.
        num_sweeps (int, optional): [description]. Defaults to 1000.
    Returns:
        list of cxxjij.utility.ClassicalSchedule, list of beta range [max, min]
    """
    if beta_min is None or beta_max is None:
        # generate Ising matrix
        ising_interaction = model.interaction_matrix()
        # set the right-bottom element zero (see issue #209)
        mat_size = ising_interaction.shape[0]
        ising_interaction[mat_size-1, mat_size-1] = 0


        if (model.vartype == openjij.BINARY):
            # convert to ising matrix
            qubo_to_ising(ising_interaction)

        abs_ising_interaction = np.abs(ising_interaction)
        max_abs_ising_interaction = np.max(abs_ising_interaction)

        #automatical setting of min, max delta energy
        abs_bias = np.sum(abs_ising_interaction, axis=1)

        #apply threshold to avoid extremely large beta_max
        THRESHOLD = 1e-8

        min_delta_energy = np.min(abs_ising_interaction[abs_ising_interaction > max_abs_ising_interaction*THRESHOLD])
        max_delta_energy = np.max(abs_bias[abs_bias > max_abs_ising_interaction*THRESHOLD])

    # TODO: More optimal schedule ?

    beta_min = np.log(2) / max_delta_energy if beta_min is None else beta_min
    beta_max = np.log(100) / min_delta_energy if beta_max is None else beta_max

    num_sweeps_per_beta = max(1, num_sweeps // 1000)

    # set schedule to cxxjij
    schedule = cxxjij.utility.make_classical_schedule_list(
        beta_min=beta_min, beta_max=beta_max,
        one_mc_step=num_sweeps_per_beta,
        num_call_updater=num_sweeps//num_sweeps_per_beta
    )

    return schedule, [beta_max, beta_min]

def geometric_hubo_beta_schedule(sa_system, beta_max, beta_min, num_sweeps):

    max_delta_energy = sa_system.get_max_effective_dE()
    min_delta_energy = sa_system.get_min_effective_dE()

    if beta_min is None:
        beta_min = np.log(2) / max_delta_energy

    if beta_max is None:
        beta_max = np.log(100) / min_delta_energy

    num_sweeps_per_beta = max(1, num_sweeps // 1000)

    schedule = cxxjij.utility.make_classical_schedule_list(
        beta_min=beta_min, beta_max=beta_max,
        one_mc_step=num_sweeps_per_beta,
        num_call_updater=num_sweeps//num_sweeps_per_beta
    )

    return schedule, [beta_max, beta_min]






# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cxxjij
import openjij
from openjij.sampler import SQASampler
from .base_gpu_chimera import BaseGPUChimeraSampler
from openjij.model import BinaryQuadraticModel, ChimeraModel
import numpy as np


class GPUChimeraSQASampler(SQASampler, BaseGPUChimeraSampler):
    """Sampler with Simulated Quantum Annealing (SQA) on GPU.

    Inherits from :class:`openjij.sampler.sqa_sampler.SQASampler`.

    Args:
        beta (float): Inverse temperature.
        gamma (float): Amplitude of quantum fluctuation.
        trotter (int): Trotter number.
        num_sweeps (int): number of sweeps
        schedule_info (dict): Information about a annealing schedule.
        num_reads (int): Number of iterations.
        unit_num_L (int): Length of one side of two-dimensional lattice in which chimera unit cells are arranged.

    Raises:
        ValueError: If variables violate as below.
        - trotter number is odd.
        - no input "unit_num_L" to an argument or this constructor.
        - given problem graph is incompatible with chimera graph.

        AttributeError: If GPU doesn't work.

    """

    def __init__(self, beta=10.0, gamma=1.0,
                 trotter=4, num_sweeps=100,
                 schedule=None, num_reads=1, unit_num_L=None):
        # GPU Sampler allows only even trotter number
        if trotter % 2 != 0:
            raise ValueError('GPU Sampler allows only even trotter number')
        self.trotter = trotter
        self.unit_num_L = unit_num_L

        super().__init__(beta=beta, gamma=gamma, trotter=trotter, 
                         num_reads=num_reads,
                         num_sweeps=num_sweeps, schedule=schedule)

        self._make_system = {
            'singlespinflip': cxxjij.system.make_chimera_transverse_gpu
        }
        self._algorithm = {
            'singlespinflip': cxxjij.algorithm.Algorithm_GPU_run
        }

    def _get_result(self, system, model):
        result = cxxjij.result.get_solution(system)
        sys_info = {}
        return result, sys_info


    def sample_ising(self, h, J,
                     beta=None, gamma=None,
                     num_sweeps=None, schedule=None,num_reads=None,
                     unit_num_L=None,
                     initial_state=None, updater=None,
                     reinitialize_state=True, seed=None):
        """Sampling from the Ising model

        Args:
            h (dict): Linear term of the target Ising model. 
            J (dict): Quadratic term of the target Ising model. 
            beta (float, optional): inverse tempareture.
            gamma (float, optional): strangth of transverse field. Defaults to None.
            num_sweeps (int, optional): number of sweeps. Defaults to None.
            schedule (list[list[float, int]], optional): List of annealing parameter. Defaults to None.
            num_reads (int, optional): number of sampling. Defaults to 1.
            initial_state (list[int], optional): Initial state. Defaults to None.
            updater (str, optional): update method. Defaults to 'single spin flip'.
            reinitialize_state (bool, optional): Re-initilization at each sampling. Defaults to True.
            seed (int, optional): Sampling seed. Defaults to None.

        Returns:
            :class:`openjij.sampler.response.Response`: results

        Examples::
            
            >>> sampler = oj.GPUChimeraSQASampler(unit_num_L=2)
            >>> h = {0: -1, 1: -1, 2: 1, 3: 1},
            >>> J = {(0, 4): -1, (2, 5): -1}
            >>> res = sampler.sample_ising(h, J)
        """

        #Set default updater
        if updater is None:
            updater = 'single spin flip'

        if num_reads is None:
            num_reads = 1

        self.unit_num_L = unit_num_L if unit_num_L else self.unit_num_L

        model = openjij.ChimeraModel(linear=h, quadratic=J, vartype='SPIN', 
                                   unit_num_L=self.unit_num_L, gpu=True)

        # define Chimera structure
        structure = {}
        structure['size'] = 8 * self.unit_num_L * self.unit_num_L
        structure['dict'] = {}
        if isinstance(model.indices[0], int):
            # identity dict
            for ind in model.indices:
                structure['dict'][ind] = ind
        elif isinstance(model.indices[0], tuple):
            # map chimera coordinate to index
            for ind in model.indices:
                structure['dict'][ind] = model.to_index(*ind, self.unit_num_L)

        return self._sampling(model, beta=beta, gamma=gamma,
                     num_sweeps=num_sweeps, schedule=schedule,
                     num_reads=num_reads,
                     initial_state=initial_state, updater=updater,
                     reinitialize_state=reinitialize_state, seed=seed, structure=structure)




import os
import re
import sys
import platform
import sysconfig
import subprocess
import shutil
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.test import test as TestCommand
from distutils.version import LooseVersion


# Package meta-data.
NAME = 'openjij'
DESCRIPTION = 'Framework for the Ising model and QUBO'
EMAIL = 'openjij@j-ij.com'
AUTHOR = 'Jij Inc.'

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.17.0':
                raise RuntimeError("CMake >= 3.17.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)
    def build_extension(self, ext):
        print(platform.system())
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        # required for auto-detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        cfg = 'Debug' if self.debug else 'Release'
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable, 
                      "-DOPENJIJ_VERSION_INFO={}".format(self.distribution.get_version()),
                      "-DCMAKE_BUILD_TYPE={}".format(cfg),  # not used on MSVC, but no harm
                      #"-DCMAKE_FIND_DEBUG_MODE=1",
                     ]
        build_args = []
        
        if platform.system() != "Windows":
            # Using Ninja-build since it a) is available as a wheel and b)
            # multithreads automatically. MSVC would require all variables be
            # exported for Ninja to pick it up, which is a little tricky to do.
            # Users can override the generator with CMAKE_GENERATOR in CMake
            # 3.15+.
            if not cmake_generator:
                try:
                    import ninja  # noqa: F401

                    cmake_args += ["-GNinja"]
                except ImportError:
                    pass

        else:

            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
                ]
                build_args += ["--config", cfg]
        
        if platform.system() == 'Darwin':
            # disable macos openmp since addtional dependency is needed.
            if not {'True': True, 'False': False}[os.getenv('FORCE_USE_OMP', 'False')]:
                print("FORCE_USE_OMP=No")
                cmake_args += ['-DFORCE_USE_OMP=No']
            else:
                print("FORCE_USE_OMP=Yes")
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
            if archs:
                cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]
                #if platform.processor() == archs: 
                #    print("COPY libomp.")
                #    shutil.copytree("/usr/local/opt/libomp", "./libomp")
            #else:
                #print("COPY libomp.")
                #shutil.copytree("/usr/local/opt/libomp", "./libomp")
                
        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += ["-j{}".format(self.parallel)]       
        if "USE_TEST" in os.environ:  
            cmake_args += ["-DUSE_TEST=Yes", "-LA", "-LH"]
        
        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''), self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)


class GoogleTestCommand(TestCommand):
    """
    A custom test runner to execute both Python unittest tests and C++ Google Tests.
    """

    def distutils_dir_name(self, dname):
        """Returns the name of a distutils build directory"""
        dir_name = "{dirname}.{platform}-{version[0]}.{version[1]}"
        return dir_name.format(dirname=dname,
                               platform=sysconfig.get_platform(),
                               version=sys.version_info)

    def run(self):
        # Run Python tests
        super(GoogleTestCommand, self).run()
        print("\nPython tests complete, now running C++ tests...\n")
        # Run catch tests
        print(os.path.join('build/', self.distutils_dir_name('lib')))
        subprocess.call(['make cxxjij_test'],
                        cwd=os.path.join('build',
                                         self.distutils_dir_name('temp')),
                        shell=True)
        subprocess.call(['./tests/cxxjij_test'],
                        cwd=os.path.join('build',
                                         self.distutils_dir_name('temp')),
                        shell=True)


class PyTestCommand(TestCommand):
    def run(self):
        super().run()


setup(
    name=NAME,
    version_config=True,
    author=AUTHOR,
    author_email='openjij@j-ij.com',
    url='https://openjij.github.io/OpenJij/',
    description='Framework for the Ising model and QUBO',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    python_requires = ">=3.7, <3.11",
    setup_requires=[
        'pytest-runner', 
        'setuptools-git-versioning', 
    ],
    install_requires=[
        'dimod >= 0.9.1',
        'numpy >= 1.18.4, !=1.21.0, !=1.21.1', 
        'scipy', 
        'requests',
        'jij-cimod >= 1.2.3', 
        'typing-extensions >= 3.10.0; python_version < "3.8.0"'
    ],
    tests_require = [
        'pytest',
        'pytest-mock',
        'pytest-cov', 
        'pytest-html',
        ],
    ext_modules=[CMakeExtension('cxxjij')],
    cmdclass=dict(build_ext=CMakeBuild, test=GoogleTestCommand,
                  pytest=PyTestCommand),
    packages=[  
        'openjij', 
        'openjij.model', 
        'openjij.sampler', 
        'openjij.sampler.chimera_gpu', 
        'openjij.utils',
        ],
    license='Apache License 2.0',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    platforms=[
                'Operating System :: Microsoft :: Windows :: Windows 10',
                'Operating System :: MacOS :: MacOS X',
                'Operating System :: POSIX :: Linux',
                ],
    zip_safe=False
)



# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module defines the abstract sampler (BaseSampler).
"""

import numpy as np
import cxxjij
import openjij
import dimod
from dimod.core.sampler import samplemixinmethod
from cimod.utils import get_state_and_energy

import time


def measure_time(func):
    """Decorator for measuring calculation time.

    Args:
        func: decorator function
    """
    def wrapper(*args, **kargs):
        start_time = time.perf_counter()

        func(*args, **kargs)

        end_time = time.perf_counter()
        execution_time = end_time - start_time
        return execution_time
    return wrapper


class BaseSampler(dimod.Sampler):
    """Base sampler class of python wrapper for cxxjij simulator

    """
    parameters = dict()
    properties = dict()

    def _set_params(self, **kwargs):
        for key, value in kwargs.items():
            if key not in self._default_params:
                raise ValueError("Unknown parameters detected")
            if value is None:
                self._params[key] = self._default_params[key]
            else:
                self._params[key] = value

    def _sampling(self, **kwargs):
        pass

    def _cxxjij_sampling(self, model, init_generator,
                         algorithm, system,
                         reinitialize_state=None,
                         seed=None, offset=None):
        """Basic sampling function: for cxxjij sampling

        Args:
            model (openjij.BinaryQuadraticModel): model has a information of instaunce (h, J, Q)
            init_generator (callable): return initial state, must have argument structure
            algorithm (callable): system algorithm of cxxjij
            system (:obj:): [description]
            reinitialize_state (bool, optional): [description]. Defaults to None.
            seed (int, optional): seed for algorithm. Defaults to None.
            offset (float): an offset which is added to the calculated energy

        Returns:
            :class:`openjij.sampler.response.Response`: results 
        """

        if offset is None:
            offset = 0

        # set algorithm function and set random seed ----
        if seed is None:
            def sampling_algorithm(system):
                return algorithm(system, self._params['schedule'])
        else:
            def sampling_algorithm(system):
                return algorithm(system, seed, self._params['schedule'])
        # ---- set algorithm function and set random seed

        # setting of response class
        execution_time = []

        # define sampling execution function ---------------
        states, energies = [], []
        system_info = {'system': []}
        @measure_time
        def exec_sampling():
            for _ in range(self._params['num_reads']):
                # Re-initialize at each sampling
                # In reverse annealing,
                # user can use previous result (if re-initilize is set to False)
                if reinitialize_state:
                    system.reset_spins(init_generator())
                # Run sampling algorithm
                # and measure execution time
                _exec_time = measure_time(sampling_algorithm)(system)
                execution_time.append(_exec_time)

                # get Ising result (state and system information)
                # ex. _sys_info save trotterized quantum state.
                result_state, _sys_info = self._get_result(system, model)

                # convert result_state to cimod style
                result_state, energy = get_state_and_energy(model, result_state, offset)

                # store result (state and energy)
                states.append(result_state)
                energies.append(energy)

                if _sys_info:
                    system_info['system'].append(_sys_info)
        # --------------- define sampling execution function

        # Execute sampling function
        sampling_time = exec_sampling()

        # construct response instance
        response = openjij.Response.from_samples(
            states, model.vartype, energies,
            info=system_info
        )

        # save execution time
        response.info['sampling_time'] = sampling_time * 10**6  # micro sec
        response.info['execution_time'] = np.mean(
            execution_time) * 10**6  # micro sec
        response.info['list_exec_times'] = np.array(
            execution_time) * 10**6  # micro sec

        return response

    def _get_result(self, system, model):
        result = cxxjij.result.get_solution(system)
        sys_info = {}
        return result, sys_info

    
    @samplemixinmethod
    def sample(self, bqm, **parameters):
        """Sample from a binary quadratic model.
        Args:
            bqm (openjij.BinaryQuadraticModel):
                Binary Qudratic Model
            **parameters:
                See the implemented sampling for additional keyword definitions.

        Returns:
            :class:`openjij.sampler.response.Response`: results 
        """
        if bqm.vartype == openjij.SPIN:
            if not getattr(self.sample_ising, '__issamplemixin__', False):
                # sample_ising is implemented
                h, J, offset = bqm.to_ising()
                sampleset = self.sample_ising(h, J, **parameters)
                sampleset.record.energy += offset
                return sampleset
            else:
                Q, offset = bqm.to_qubo()
                sampleset = self.sample_qubo(Q, **parameters)
                sampleset.change_vartype(dimod.SPIN, energy_offset=offset)
                return sampleset
        elif bqm.vartype == openjij.BINARY:
            if not getattr(self.sample_qubo, '__issamplemixin__', False):
                # sample_qubo is implemented
                Q, offset = bqm.to_qubo()
                sampleset = self.sample_qubo(Q, **parameters)
                sampleset.record.energy += offset
                return sampleset
            else:
                h, J, offset = bqm.to_ising()
                sampleset = self.sample_ising(h, J, **parameters)
                sampleset.change_vartype(dimod.BINARY, energy_offset=offset)
                return sampleset
        else:
            raise RuntimeError("binary quadratic model has an unknown vartype")

    

    @samplemixinmethod
    def sample_ising(self, h, J, **parameters):
        """Sample from an Ising model using the implemented sample method.

        Args:
            h (dict): Linear biases
            J (dict): Quadratic biases

        Returns:
            :class:`openjij.sampler.response.Response`: results 
        """
        bqm = openjij.BinaryQuadraticModel.from_ising(h, J, sparse=parameters.get('sparse', False))
        return self.sample(bqm, **parameters)


    @samplemixinmethod
    def sample_qubo(self, Q, **parameters):
        """Sample from a QUBO model using the implemented sample method.

        Args:
            Q (dict or numpy.ndarray): Coefficients of a quadratic unconstrained binary optimization

        Returns:
            :class:`openjij.sampler.response.Response`: results 
        """
        if isinstance(Q, dict):
            bqm = openjij.BinaryQuadraticModel.from_qubo(Q, sparse=parameters.get('sparse', False))
            return self.sample(bqm, **parameters)
        elif isinstance(Q, np.ndarray):
            bqm = openjij.BinaryQuadraticModel.from_numpy_matrix(Q, vartype='BINARY')
            return self.sample(bqm, **parameters)
        else:
            raise TypeError('Q must be either dict or np.ndarray')




from .benchmark import solver_benchmark, residual_energy, time_to_solution, success_probability, se_residual_energy, se_success_probability, se_upper_tts, se_lower_tts
from .res_convertor import convert_response
import openjij.utils.graph_utils



# Copyright 2021 Jij Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import openjij

def convert_response(response):

    if isinstance(response, openjij.Response):
        return response

    try:
        from dimod.response import Response
        from dimod.sampleset import SampleSet
    except ImportError:
        raise ImportError('Import dwave dimod : "pip install dimod"')
    
    if isinstance(response, Response) or isinstance(response, SampleSet):
        from dimod.vartypes import BINARY, SPIN
        vartype = 'BINARY' if response.vartype == BINARY else 'SPIN'
        o_res = openjij.Response(vartype=vartype, indices=list(response.variables))
        states = []
        energies = []
        for rec in response.record:
            for _ in range(rec[2]):
                states.append(rec[0])
                energies.append(rec[1])
        o_res.update_ising_states_energies(states=states, energies=energies)
                
        o_res.info = response.info
        if 'qpu_sampling_time' in response.info:
            o_res.info['sampling_time'] = response.info['qpu_sampling_time']
        if 'anneal_time_per_run' in response.info:
            o_res.info['execution_time'] = response.info['anneal_time_per_run']
        o_res.info['dimod'] = response
        return o_res
    else:
        raise TypeError('response is dimod.response.Response, SampleSet or openjij.Response')
